from textwrap import dedent
from enum import Enum, EnumMeta
from typing import Optional, Union, Any, get_args

from .instance import connects, connectable
from .signal import Signal


@connects
@connectable
class InterfaceInstance:
    """ Instance of an Interface, 
    Generally in a Module or another Interface """

    _specialcases = [
        "name",
        "of",
        "port",
        "role",
        "src",
        "dest",
        "conns",
        "portrefs",
        "_elaborated",
        "_initialized",
    ]

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        of: "Interface",
        port: bool = False,
        role: Optional[Enum] = None,
        src: Optional[Enum] = None,
        dest: Optional[Enum] = None,
    ):
        self.name = name
        self.of = of
        self.port = port
        self.role = role
        self.src = src
        self.dest = dest
        self.conns = dict()
        self.portrefs = dict()
        self._elaborated = False
        self._initialized = True

    @property
    def _resolved(self) -> "Interface":
        return self.of


# Type-alias for HDL objects storable as `Module` attributes
InterfaceAttr = Union[Signal, InterfaceInstance]


def _is_interface_attr(val: Any) -> bool:
    """ Boolean indication of whether `val` is a valid `hdl21.Interface` attribute. """
    return isinstance(val, get_args(InterfaceAttr))


class Interface:
    """ 
    # hdl21 Interface
    
    Interfaces are structured hierarchical connection objects which include Signals and other Interfaces. 
    """

    def __init__(self, *, name=None):
        self.name = name
        self.roles = None
        self.signals = dict()
        self.interfaces = dict()
        self.namespace = dict()  # Combination of all these
        self._elaborated = False
        self._initialized = True

    def __setattr__(self, key: str, val: object):
        """ Set-attribute over-ride, organizing into type-based containers """
        from .signal import Signal

        if not getattr(self, "_initialized", False) or key.startswith("_"):
            return super().__setattr__(key, val)

        # Protected attrs - the internal dicts
        banned = ["signals", "interfaces", "namespace"]
        if key in banned:
            raise RuntimeError(
                f"Error attempting to over-write protected attribute {key} of Interface {self}"
            )
        # Special case(s)
        if key == "name":
            return super().__setattr__(key, val)
        if key == "roles":
            if not isinstance(val, EnumMeta):
                raise TypeError(f"Interface roles must be an `enum.Enum`")
            return super().__setattr__(key, val)

        # Type-based organization
        if isinstance(val, Signal):
            val.name = key
            self.namespace[key] = val
            self.signals[key] = val
        elif isinstance(val, InterfaceInstance):
            val.name = key
            self.interfaces[key] = val
            self.namespace[key] = val
        else:
            raise TypeError(f"Invalid Interface attribute {val} for {self}")

    def __getattr__(self, key):
        ns = self.__getattribute__("namespace")
        if key in ns:
            return ns[key]
        return object.__getattribute__(self, key)

    def __call__(self, **kwargs):
        """ Calls to Interfaces return Interface Instances """
        return InterfaceInstance(of=self, **kwargs)

    def __init_subclass__(cls, *_args, **_kwargs):
        """ Sub-Classing Disable-ization """
        raise RuntimeError(
            dedent(
                f"""\
                Error attempting to create {cls.__name__}
                Sub-Typing hdl21.Interface is not supported. """
            )
        )

    def add(self, val: InterfaceAttr, *, name: Optional[str] = None) -> InterfaceAttr:
        raise NotImplementedError

    def get(self, name: str) -> Optional[InterfaceAttr]:
        """ Get module-attribute `name`. Returns `None` if not present. 
        Note unlike Python built-ins such as `getattr`, `get` returns solely 
        from the HDL namespace-worth of `InterfaceAttr`s. """
        ns = self.__getattribute__("namespace")
        return ns.get(name, None)

    @property
    def Roles(self):
        # Roles often look like a class, so they have a class-style name-accessor
        return self.roles


def interface(cls: type) -> Interface:
    """ # Interface Definition Decorator 
    
    Converts a class-body full of Interface-storable attributes (Signals, other Interfaces) to an `hdl21.Interface`. 
    Example Usage: 

    ```python
    import hdl21 as h

    @h.interface
    class Diff:
        p = h.Signal()
        n = h.Signal()

    @h.interface
    class DisplayPort:
        main_link = Diff()
        aux = h.Signal()
    ```

    Interfaces may also define a `Roles` enumeration, inline within this class-body. 
    `Roles` are optional pieces of enumerated endpoint-labels which aid in dictating `Signal` directions. 
    Each `Signal` accepts optional source (`src`) and destination (`dst`) fields which (if set) must be one of these roles. 
    
    ```python
    import hdl21 as h
    from enum import Enum, auto

    @h.interface
    class HasRoles:
        class Roles(Enum):
            HOST = auto()
            DEVICE = auto()

        tx = h.Signal(src=Roles.HOST, dest=Roles.DEVICE)
        rx = h.Signal(src=Roles.DEVICE, dest=Roles.HOST)
    ```

    """
    if cls.__bases__ != (object,):
        raise RuntimeError(f"Invalid @hdl21.interface inheriting from {cls.__bases__}")

    # Create the Interface object
    intf = Interface(name=cls.__name__)

    protected_names = ["signals", "interfaces"]
    dunders = dict()
    unders = dict()

    # Take a lap through the class dictionary, type-check everything and assign relevant attributes to the interface
    for key, val in cls.__dict__.items():
        if key in protected_names:
            raise RuntimeError(f"Invalid field name {key} in interface {cls}")
        elif key.startswith("__"):
            dunders[key] = val
        elif key.startswith("_"):
            unders[key] = val
        elif key == "Roles":
            # Special-case the upper-cased `Roles`, as it'll often be a class-def
            setattr(intf, "roles", val)
        else:
            setattr(intf, key, val)
    # And return the interface
    return intf



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: circuit.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='circuit.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\rcircuit.proto\"\x7f\n\x07Package\x12\x0e\n\x06\x64omain\x18\x01 \x01(\t\x12\x18\n\x07modules\x18\x02 \x03(\x0b\x32\x07.Module\x12$\n\x0b\x65xt_modules\x18\x03 \x03(\x0b\x32\x0f.ExternalModule\x12$\n\x0b\x65xt_sources\x18\x04 \x03(\x0b\x32\x0f.ExternalSource\"-\n\rQualifiedName\x12\x0e\n\x06\x64omain\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"F\n\tReference\x12\x0f\n\x05local\x18\x01 \x01(\tH\x00\x12\"\n\x08\x65xternal\x18\x02 \x01(\x0b\x32\x0e.QualifiedNameH\x00\x42\x04\n\x02to\"^\n\tParameter\x12\x11\n\x07integer\x18\x02 \x01(\x03H\x00\x12\x10\n\x06\x64ouble\x18\x03 \x01(\x01H\x00\x12\x10\n\x06string\x18\x04 \x01(\tH\x00\x12\x11\n\x07literal\x18\x05 \x01(\tH\x00\x42\x07\n\x05value\"|\n\x04Port\x12\x17\n\x06signal\x18\x01 \x01(\x0b\x32\x07.Signal\x12\"\n\tdirection\x18\x02 \x01(\x0e\x32\x0f.Port.Direction\"7\n\tDirection\x12\t\n\x05INPUT\x10\x00\x12\n\n\x06OUTPUT\x10\x01\x12\t\n\x05INOUT\x10\x02\x12\x08\n\x04NONE\x10\x03\"%\n\x06Signal\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05width\x18\x02 \x01(\x03\"1\n\x05Slice\x12\x0e\n\x06signal\x18\x01 \x01(\t\x12\x0b\n\x03top\x18\x02 \x01(\x03\x12\x0b\n\x03\x62ot\x18\x03 \x01(\x03\"$\n\x06\x43oncat\x12\x1a\n\x05parts\x18\x01 \x03(\x0b\x32\x0b.Connection\"a\n\nConnection\x12\x16\n\x03sig\x18\x01 \x01(\x0b\x32\x07.SignalH\x00\x12\x17\n\x05slice\x18\x02 \x01(\x0b\x32\x06.SliceH\x00\x12\x19\n\x06\x63oncat\x18\x03 \x01(\x0b\x32\x07.ConcatH\x00\x42\x07\n\x05stype\"\x94\x02\n\x08Instance\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x06module\x18\x02 \x01(\x0b\x32\n.Reference\x12-\n\nparameters\x18\x03 \x03(\x0b\x32\x19.Instance.ParametersEntry\x12/\n\x0b\x63onnections\x18\x04 \x03(\x0b\x32\x1a.Instance.ConnectionsEntry\x1a=\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x19\n\x05value\x18\x02 \x01(\x0b\x32\n.Parameter:\x02\x38\x01\x1a?\n\x10\x43onnectionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1a\n\x05value\x18\x02 \x01(\x0b\x32\x0b.Connection:\x02\x38\x01\"\xe6\x01\n\x06Module\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x05ports\x18\x02 \x03(\x0b\x32\x05.Port\x12\x18\n\x07signals\x18\x03 \x03(\x0b\x32\x07.Signal\x12\x1c\n\tinstances\x18\x04 \x03(\x0b\x32\t.Instance\x12:\n\x12\x64\x65\x66\x61ult_parameters\x18\x05 \x03(\x0b\x32\x1e.Module.DefaultParametersEntry\x1a\x44\n\x16\x44\x65\x66\x61ultParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x19\n\x05value\x18\x02 \x01(\x0b\x32\n.Parameter:\x02\x38\x01\"R\n\x0e\x45xternalModule\x12\x1c\n\x04name\x18\x01 \x01(\x0b\x32\x0e.QualifiedName\x12\x14\n\x05ports\x18\x02 \x03(\x0b\x32\x05.Port\x12\x0c\n\x04\x64\x65sc\x18\x03 \x01(\t\"6\n\x0e\x45xternalSource\x12\x0e\n\x04path\x18\x01 \x01(\tH\x00\x12\r\n\x03url\x18\x02 \x01(\tH\x00\x42\x05\n\x03srcb\x06proto3'
)



_PORT_DIRECTION = _descriptor.EnumDescriptor(
  name='Direction',
  full_name='Port.Direction',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INPUT', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='OUTPUT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INOUT', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NONE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=430,
  serialized_end=485,
)
_sym_db.RegisterEnumDescriptor(_PORT_DIRECTION)


_PACKAGE = _descriptor.Descriptor(
  name='Package',
  full_name='Package',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='domain', full_name='Package.domain', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='modules', full_name='Package.modules', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ext_modules', full_name='Package.ext_modules', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ext_sources', full_name='Package.ext_sources', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17,
  serialized_end=144,
)


_QUALIFIEDNAME = _descriptor.Descriptor(
  name='QualifiedName',
  full_name='QualifiedName',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='domain', full_name='QualifiedName.domain', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='QualifiedName.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=146,
  serialized_end=191,
)


_REFERENCE = _descriptor.Descriptor(
  name='Reference',
  full_name='Reference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='local', full_name='Reference.local', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='external', full_name='Reference.external', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='to', full_name='Reference.to',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=193,
  serialized_end=263,
)


_PARAMETER = _descriptor.Descriptor(
  name='Parameter',
  full_name='Parameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='integer', full_name='Parameter.integer', index=0,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='double', full_name='Parameter.double', index=1,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='string', full_name='Parameter.string', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='literal', full_name='Parameter.literal', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='Parameter.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=265,
  serialized_end=359,
)


_PORT = _descriptor.Descriptor(
  name='Port',
  full_name='Port',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signal', full_name='Port.signal', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='direction', full_name='Port.direction', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PORT_DIRECTION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=361,
  serialized_end=485,
)


_SIGNAL = _descriptor.Descriptor(
  name='Signal',
  full_name='Signal',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Signal.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='width', full_name='Signal.width', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=487,
  serialized_end=524,
)


_SLICE = _descriptor.Descriptor(
  name='Slice',
  full_name='Slice',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signal', full_name='Slice.signal', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='top', full_name='Slice.top', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bot', full_name='Slice.bot', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=526,
  serialized_end=575,
)


_CONCAT = _descriptor.Descriptor(
  name='Concat',
  full_name='Concat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parts', full_name='Concat.parts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=577,
  serialized_end=613,
)


_CONNECTION = _descriptor.Descriptor(
  name='Connection',
  full_name='Connection',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sig', full_name='Connection.sig', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='slice', full_name='Connection.slice', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='concat', full_name='Connection.concat', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='stype', full_name='Connection.stype',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=615,
  serialized_end=712,
)


_INSTANCE_PARAMETERSENTRY = _descriptor.Descriptor(
  name='ParametersEntry',
  full_name='Instance.ParametersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Instance.ParametersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Instance.ParametersEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=865,
  serialized_end=926,
)

_INSTANCE_CONNECTIONSENTRY = _descriptor.Descriptor(
  name='ConnectionsEntry',
  full_name='Instance.ConnectionsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Instance.ConnectionsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Instance.ConnectionsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=928,
  serialized_end=991,
)

_INSTANCE = _descriptor.Descriptor(
  name='Instance',
  full_name='Instance',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Instance.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='module', full_name='Instance.module', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='Instance.parameters', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='connections', full_name='Instance.connections', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_INSTANCE_PARAMETERSENTRY, _INSTANCE_CONNECTIONSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=715,
  serialized_end=991,
)


_MODULE_DEFAULTPARAMETERSENTRY = _descriptor.Descriptor(
  name='DefaultParametersEntry',
  full_name='Module.DefaultParametersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Module.DefaultParametersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Module.DefaultParametersEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1156,
  serialized_end=1224,
)

_MODULE = _descriptor.Descriptor(
  name='Module',
  full_name='Module',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Module.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ports', full_name='Module.ports', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signals', full_name='Module.signals', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='instances', full_name='Module.instances', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='default_parameters', full_name='Module.default_parameters', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MODULE_DEFAULTPARAMETERSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=994,
  serialized_end=1224,
)


_EXTERNALMODULE = _descriptor.Descriptor(
  name='ExternalModule',
  full_name='ExternalModule',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='ExternalModule.name', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ports', full_name='ExternalModule.ports', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='desc', full_name='ExternalModule.desc', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1226,
  serialized_end=1308,
)


_EXTERNALSOURCE = _descriptor.Descriptor(
  name='ExternalSource',
  full_name='ExternalSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='ExternalSource.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='ExternalSource.url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='src', full_name='ExternalSource.src',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1310,
  serialized_end=1364,
)

_PACKAGE.fields_by_name['modules'].message_type = _MODULE
_PACKAGE.fields_by_name['ext_modules'].message_type = _EXTERNALMODULE
_PACKAGE.fields_by_name['ext_sources'].message_type = _EXTERNALSOURCE
_REFERENCE.fields_by_name['external'].message_type = _QUALIFIEDNAME
_REFERENCE.oneofs_by_name['to'].fields.append(
  _REFERENCE.fields_by_name['local'])
_REFERENCE.fields_by_name['local'].containing_oneof = _REFERENCE.oneofs_by_name['to']
_REFERENCE.oneofs_by_name['to'].fields.append(
  _REFERENCE.fields_by_name['external'])
_REFERENCE.fields_by_name['external'].containing_oneof = _REFERENCE.oneofs_by_name['to']
_PARAMETER.oneofs_by_name['value'].fields.append(
  _PARAMETER.fields_by_name['integer'])
_PARAMETER.fields_by_name['integer'].containing_oneof = _PARAMETER.oneofs_by_name['value']
_PARAMETER.oneofs_by_name['value'].fields.append(
  _PARAMETER.fields_by_name['double'])
_PARAMETER.fields_by_name['double'].containing_oneof = _PARAMETER.oneofs_by_name['value']
_PARAMETER.oneofs_by_name['value'].fields.append(
  _PARAMETER.fields_by_name['string'])
_PARAMETER.fields_by_name['string'].containing_oneof = _PARAMETER.oneofs_by_name['value']
_PARAMETER.oneofs_by_name['value'].fields.append(
  _PARAMETER.fields_by_name['literal'])
_PARAMETER.fields_by_name['literal'].containing_oneof = _PARAMETER.oneofs_by_name['value']
_PORT.fields_by_name['signal'].message_type = _SIGNAL
_PORT.fields_by_name['direction'].enum_type = _PORT_DIRECTION
_PORT_DIRECTION.containing_type = _PORT
_CONCAT.fields_by_name['parts'].message_type = _CONNECTION
_CONNECTION.fields_by_name['sig'].message_type = _SIGNAL
_CONNECTION.fields_by_name['slice'].message_type = _SLICE
_CONNECTION.fields_by_name['concat'].message_type = _CONCAT
_CONNECTION.oneofs_by_name['stype'].fields.append(
  _CONNECTION.fields_by_name['sig'])
_CONNECTION.fields_by_name['sig'].containing_oneof = _CONNECTION.oneofs_by_name['stype']
_CONNECTION.oneofs_by_name['stype'].fields.append(
  _CONNECTION.fields_by_name['slice'])
_CONNECTION.fields_by_name['slice'].containing_oneof = _CONNECTION.oneofs_by_name['stype']
_CONNECTION.oneofs_by_name['stype'].fields.append(
  _CONNECTION.fields_by_name['concat'])
_CONNECTION.fields_by_name['concat'].containing_oneof = _CONNECTION.oneofs_by_name['stype']
_INSTANCE_PARAMETERSENTRY.fields_by_name['value'].message_type = _PARAMETER
_INSTANCE_PARAMETERSENTRY.containing_type = _INSTANCE
_INSTANCE_CONNECTIONSENTRY.fields_by_name['value'].message_type = _CONNECTION
_INSTANCE_CONNECTIONSENTRY.containing_type = _INSTANCE
_INSTANCE.fields_by_name['module'].message_type = _REFERENCE
_INSTANCE.fields_by_name['parameters'].message_type = _INSTANCE_PARAMETERSENTRY
_INSTANCE.fields_by_name['connections'].message_type = _INSTANCE_CONNECTIONSENTRY
_MODULE_DEFAULTPARAMETERSENTRY.fields_by_name['value'].message_type = _PARAMETER
_MODULE_DEFAULTPARAMETERSENTRY.containing_type = _MODULE
_MODULE.fields_by_name['ports'].message_type = _PORT
_MODULE.fields_by_name['signals'].message_type = _SIGNAL
_MODULE.fields_by_name['instances'].message_type = _INSTANCE
_MODULE.fields_by_name['default_parameters'].message_type = _MODULE_DEFAULTPARAMETERSENTRY
_EXTERNALMODULE.fields_by_name['name'].message_type = _QUALIFIEDNAME
_EXTERNALMODULE.fields_by_name['ports'].message_type = _PORT
_EXTERNALSOURCE.oneofs_by_name['src'].fields.append(
  _EXTERNALSOURCE.fields_by_name['path'])
_EXTERNALSOURCE.fields_by_name['path'].containing_oneof = _EXTERNALSOURCE.oneofs_by_name['src']
_EXTERNALSOURCE.oneofs_by_name['src'].fields.append(
  _EXTERNALSOURCE.fields_by_name['url'])
_EXTERNALSOURCE.fields_by_name['url'].containing_oneof = _EXTERNALSOURCE.oneofs_by_name['src']
DESCRIPTOR.message_types_by_name['Package'] = _PACKAGE
DESCRIPTOR.message_types_by_name['QualifiedName'] = _QUALIFIEDNAME
DESCRIPTOR.message_types_by_name['Reference'] = _REFERENCE
DESCRIPTOR.message_types_by_name['Parameter'] = _PARAMETER
DESCRIPTOR.message_types_by_name['Port'] = _PORT
DESCRIPTOR.message_types_by_name['Signal'] = _SIGNAL
DESCRIPTOR.message_types_by_name['Slice'] = _SLICE
DESCRIPTOR.message_types_by_name['Concat'] = _CONCAT
DESCRIPTOR.message_types_by_name['Connection'] = _CONNECTION
DESCRIPTOR.message_types_by_name['Instance'] = _INSTANCE
DESCRIPTOR.message_types_by_name['Module'] = _MODULE
DESCRIPTOR.message_types_by_name['ExternalModule'] = _EXTERNALMODULE
DESCRIPTOR.message_types_by_name['ExternalSource'] = _EXTERNALSOURCE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Package = _reflection.GeneratedProtocolMessageType('Package', (_message.Message,), {
  'DESCRIPTOR' : _PACKAGE,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Package)
  })
_sym_db.RegisterMessage(Package)

QualifiedName = _reflection.GeneratedProtocolMessageType('QualifiedName', (_message.Message,), {
  'DESCRIPTOR' : _QUALIFIEDNAME,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:QualifiedName)
  })
_sym_db.RegisterMessage(QualifiedName)

Reference = _reflection.GeneratedProtocolMessageType('Reference', (_message.Message,), {
  'DESCRIPTOR' : _REFERENCE,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Reference)
  })
_sym_db.RegisterMessage(Reference)

Parameter = _reflection.GeneratedProtocolMessageType('Parameter', (_message.Message,), {
  'DESCRIPTOR' : _PARAMETER,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Parameter)
  })
_sym_db.RegisterMessage(Parameter)

Port = _reflection.GeneratedProtocolMessageType('Port', (_message.Message,), {
  'DESCRIPTOR' : _PORT,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Port)
  })
_sym_db.RegisterMessage(Port)

Signal = _reflection.GeneratedProtocolMessageType('Signal', (_message.Message,), {
  'DESCRIPTOR' : _SIGNAL,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Signal)
  })
_sym_db.RegisterMessage(Signal)

Slice = _reflection.GeneratedProtocolMessageType('Slice', (_message.Message,), {
  'DESCRIPTOR' : _SLICE,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Slice)
  })
_sym_db.RegisterMessage(Slice)

Concat = _reflection.GeneratedProtocolMessageType('Concat', (_message.Message,), {
  'DESCRIPTOR' : _CONCAT,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Concat)
  })
_sym_db.RegisterMessage(Concat)

Connection = _reflection.GeneratedProtocolMessageType('Connection', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTION,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Connection)
  })
_sym_db.RegisterMessage(Connection)

Instance = _reflection.GeneratedProtocolMessageType('Instance', (_message.Message,), {

  'ParametersEntry' : _reflection.GeneratedProtocolMessageType('ParametersEntry', (_message.Message,), {
    'DESCRIPTOR' : _INSTANCE_PARAMETERSENTRY,
    '__module__' : 'circuit_pb2'
    # @@protoc_insertion_point(class_scope:Instance.ParametersEntry)
    })
  ,

  'ConnectionsEntry' : _reflection.GeneratedProtocolMessageType('ConnectionsEntry', (_message.Message,), {
    'DESCRIPTOR' : _INSTANCE_CONNECTIONSENTRY,
    '__module__' : 'circuit_pb2'
    # @@protoc_insertion_point(class_scope:Instance.ConnectionsEntry)
    })
  ,
  'DESCRIPTOR' : _INSTANCE,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Instance)
  })
_sym_db.RegisterMessage(Instance)
_sym_db.RegisterMessage(Instance.ParametersEntry)
_sym_db.RegisterMessage(Instance.ConnectionsEntry)

Module = _reflection.GeneratedProtocolMessageType('Module', (_message.Message,), {

  'DefaultParametersEntry' : _reflection.GeneratedProtocolMessageType('DefaultParametersEntry', (_message.Message,), {
    'DESCRIPTOR' : _MODULE_DEFAULTPARAMETERSENTRY,
    '__module__' : 'circuit_pb2'
    # @@protoc_insertion_point(class_scope:Module.DefaultParametersEntry)
    })
  ,
  'DESCRIPTOR' : _MODULE,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:Module)
  })
_sym_db.RegisterMessage(Module)
_sym_db.RegisterMessage(Module.DefaultParametersEntry)

ExternalModule = _reflection.GeneratedProtocolMessageType('ExternalModule', (_message.Message,), {
  'DESCRIPTOR' : _EXTERNALMODULE,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:ExternalModule)
  })
_sym_db.RegisterMessage(ExternalModule)

ExternalSource = _reflection.GeneratedProtocolMessageType('ExternalSource', (_message.Message,), {
  'DESCRIPTOR' : _EXTERNALSOURCE,
  '__module__' : 'circuit_pb2'
  # @@protoc_insertion_point(class_scope:ExternalSource)
  })
_sym_db.RegisterMessage(ExternalSource)


_INSTANCE_PARAMETERSENTRY._options = None
_INSTANCE_CONNECTIONSENTRY._options = None
_MODULE_DEFAULTPARAMETERSENTRY._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hdl21', 'hdl21.netlist', 'hdl21.proto', 'hdl21.tests']

package_data = \
{'': ['*']}

install_requires = \
['protobuf>=3.17.0,<4.0.0', 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'hdl21',
    'version': '0.1.1',
    'description': 'Hardware Description Library',
    'long_description': '# HDL21\n\n## Generator-Based Structural Hardware Description Library\n\n[![test](https://github.com/dan-fritchman/Hdl21/actions/workflows/test.yaml/badge.svg)](https://github.com/dan-fritchman/Hdl21/actions/workflows/test.yaml)\n[![codecov](https://codecov.io/gh/dan-fritchman/Hdl21/branch/main/graph/badge.svg?token=f8LKUqEPdq)](https://codecov.io/gh/dan-fritchman/Hdl21)\n\nHdl21 is a library for efficiently creating and manipulating structural hardware descriptions such as those common in custom integrated circuits. Circuits are described in two primary units of re-use: \n\n- `Modules` are structural combinations of ports, signals, and instances of other `Modules`. \n- `Generator`s are Python functions that return `Modules`. \n\nIn other words: \n\n- `Generators` are code. `Modules` are data. \n- `Generators` require a runtime environment. `Modules` do not. \n\n## Modules\n\nHdl21 `Modules` are type-specific containers of just a handful of `hdl21` types. They can include: \n\n* Instances of other `Modules`\n* Connections between them defined by `Signals` and `Ports`\n* Structured connections defined by `Interfaces` \n\nAn example of creating a `Module`:\n\n```python\nimport hdl21 as h \n\nm = h.Module(name="MyModule")\nm.i = h.Input()\nm.o = h.Output(width=8)\nm.s = h.Signal()\nm.a = AnotherModule()\n```\n\nIn addition to the procedural-syntax shown above, `Modules` can also be defined through a `class`-based syntax by applying the `hdl21.module` decorator to a class-definition. \n\n```python\nimport hdl21 as h \n\n@h.module\nclass MyModule:\n    i = h.Input()\n    o = h.Output(width=8)\n    s = h.Signal()\n    a = AnotherModule()\n```\n\nThis class-based syntax produces identical results to the procedural code-block above. Its declarative style can be much more natural and expressive in many contexts, including for designers familiar with popular HDLs. \n\nCreation of `Module` signal-attributes is generally performed by the built-in `Signal`, `Port`, `Input`, and `Output` constructors. Each comes with a "plural version" (`Input*s*` etc.) which creates several identical objects at once:\n\n```python\nimport hdl21 as h\n\n@h.module\nclass MyModule:\n    a, b = h.Inputs(2)\n    c, d, e = h.Outputs(3, width=16)\n    f, g, h, i = h.Signals(4)\n```\n\n## Connections \n\nPopular HDLs generally feature one of two forms of connection semantics. Verilog, VHDL, and most dedicated HDLs use "connect by call" semantics, in which signal-objects are first declared, then passed as function-call-style arguments to instances of other modules. \n\n```verilog\nmodule my_module();\n  logic a, b, c;                              // Declare signals \n  another_module i1 (a, b, c);                // Create an instance\n  another_module i2 (.a(a), .b(b), .c(c));    // Another instance, connected by-name\nendmodule\n```\n\nChisel, in contrast, uses "connection by assignment" - more literally using the walrus `:=` operator. Instances of child modules are created first, and their ports are directly walrus-connected to one another. No local-signal objects ever need be declared in the instantiating parent module. \n\n```scala\nclass MyModule extends Module {\n  // Create Module Instances\n  val i1 = Module(new AnotherModule)\n  val i2 = Module(new AnotherModule)\n  // Wire them directly to one another \n  i1.io.a := i2.io.a\n  i1.io.b := i2.io.b\n  i1.io.c := i2.io.c\n}\n```\n\nEach can be more concise and expressive depending on context. Hdl21 `Modules` support **both** connect-by-call and connect-by-assignment forms. \n\nConnections-by-assignment are performed by assigning either a `Signal` or another instance\'s `Port` to an attribute of a Module-Instance. \n\n```python \n# Create a module\nm = h.Module()\n# Create its internal Signals\nm.a = Signal()\nm.b = Signal()\nm.c = Signal()\n# Create an Instance\nm.i1 = AnotherModule()\n# And wire them up\nm.i1.a = a\nm.i1.b = b\nm.i1.c = c\n```\n\nThis also works without the parent-module `Signals`:\n\n```python\n# Create a module\nm = h.Module() \n# Create the Instances\nm.i1 = AnotherModule()\nm.i2 = AnotherModule()\n# And wire them up\nm.i1.a = m.i2.a\nm.i1.b = m.i2.b\nm.i1.c = m.i2.c\n```\n\nInstances can instead be connected by call:\n\n```python\n# Create a module\nm = h.Module() \n# Create the Instances\nm.i1 = AnotherModule()\nm.i2 = AnotherModule()\n# Call one to connect them\nm.i1(a=m.i2.a, b=m.i2.b, c=m.i2.c)\n```\n\nThese connection-calls can also be performed inline, as the instances are being created. \n\n```python\n# Create a module\nm = h.Module() \n# Create the Instance `i1`\nm.i1 = AnotherModule()\n# Create another Instance `i2`, and connect to `i1`\nm.i2 = AnotherModule(a=m.i1.a, b=m.i1.b, c=m.i1.c)\n```\n\n\n## Generators\n\nHdl21 `Modules` are "plain old data". They require no runtime or execution environment. They can be (and are!) fully represented in markup languages such as JSON, YAML, or ProtoBuf. The power of embedding `Modules` in a general-purpose programming language lies in allowing code to create and manipulate them. Hdl21\'s `Generators` are functions which produce `Modules`, and have a number of built-in features to aid embedding in a hierarchical hardware tree. \n\nCreating a generator just requires applying the `@hdl21.generator` decorator to a function: \n\n```python\nimport hdl21 as h \n\n@h.generator\ndef MyFirstGenerator(params: MyParams) -> h.Module:\n    # A very exciting first generator function \n    m = h.Module()\n    m.i = h.Input(width=params.w)\n    return m\n```\n\nThe generator-function body can define a `Module` however it likes - procedurally or via the class-style syntax. \n\n```python\n@h.generator\ndef MySecondGenerator(params: MyParams) -> h.Module:\n    # A very exciting (second) generator function \n    @h.module\n    class MySecondGen:\n        i = h.Input(width=params.w)\n    return MySecondGen\n```\n\nOr any combination of the two:\n\n```python\n@h.generator\ndef MyThirdGenerator(params: MyParams) -> h.Module:\n    # Create an internal Module\n    @h.module\n    class Inner:\n        i = h.Input(width=params.w)\n\n    # Manipulate it a bit\n    Inner.o = h.Output(width=2 * Inner.i.width)\n\n    # Instantiate that in another Module \n    @h.module\n    class Outer:\n        inner = Inner()\n\n    # And manipulate that some more too \n    Outer.inp = h.Input(width=params.w)\n    return Outer\n```\n\nNote in the latter example, the Module `Inner` is defined solely inside the generator-function body. It serves as a local, semi-private implementation space for the returned `Outer` module. The desire for these relatively hidden implementation-details are fairly common for hierarchical hardware designs. (Sorting out which Modules are designed to be used "publicly" is a common problem.) "Closure Modules" such as `Inner` are about as private such spaces as the Python language allows. \n\n\n## Parameters \n\n`Generators` typically take a single argument `params` which is a collection of `hdl21.Params`. Generator parameters are strongly type-checked at runtime. Each requires a data-type `dtype` and description-string `desc`. Optional parameters include a default-value, which must be an instance of `dtype`.\n\n```python\nnpar = h.Param(dtype=int, desc="Number of parallel fingers", default=1)\n```\n\nThe collections of these parameters used by `Generators` are called param-classes, and are typically formed by applying the `hdl21.paramclass` decorator to a class-body-full of `hdl21.Params`: \n\n```python\nimport hdl21 as h\n\n@h.paramclass\nclass MyParams:\n    # Required\n    width = h.Param(dtype=int, desc="Input bit-width. Required")\n    # Optional - including a default value\n    text = h.Param(dtype=str, desc="An Optional string-valued parameter", default="My Favorite Module")\n```\n\nEach param-class is defined similarly to the Python standard-library\'s `dataclass`, but using assignment rather than type-annotation syntax. The `paramclass` decorator converts these class-definitions into type-checked `dataclasses`, with fields using the `dtype` of each parameter. \n\n```python\np = MyParams(width=8, text="Your Favorite Module")\nassert p.width == 8  # Passes. Note this is an `int`, not a `Param`\nassert p.text == "Your Favorite Module"  # Also passes \n```\n \nThe `paramclass` decorator also adds `descriptions` and `defaults` methods which provide dictionaries of each parameter\'s default-values and string-descriptions. Similar to `dataclasses`, param-class constructors use the field-order defined in the class body. Note Python\'s function-argument rules dictate that all required arguments be declared first, and all optional arguments come last. \n\nParam-classes can be nested, and can be converted to (potentially nested) dictionaries via `dataclasses.asdict`. The same conversion applies in reverse - (potentially nested) dictionaries can be expanded to serve as param-class constructor arguments: \n\n```python \nimport hdl21 as h\nfrom dataclasses import asdict\n\n@h.paramclass\nclass Inner:\n    i = h.Param(dtype=int, desc="Inner int-field")\n\n@h.paramclass\nclass Outer:\n    inner = h.Param(dtype=Inner, desc="Inner fields")\n    f = h.Param(dtype=float, desc="A float", default=3.14159)\n\n# Create from a (nested) dictionary literal \nd1 = {"inner": {"i": 11}, "f": 22.2}\no = Outer(**d1)\n# Convert back to another dictionary \nd2 = asdict(o)\n# And check they line up \nassert d1 == d2\n```\n\nParam-classes are immutable and hashable, and therefore require their attributes to be as well. Two common types left out by these rules are `list` and `dict`. Lists are easily converted into tuples. Param-class creation performs these conversions inline: \n\n```python\n@h.paramclass\nclass HasTuple:\n    t = h.Param(dtype=tuple, desc="Go ahead, try a list")\n\nh = HasTuple(t=[1, 2, 3])      # Give it a list\nassert isinstance(h.t, tuple)  # Passes\nassert h.t == (1, 2, 3)        # Also passes \n```\n\nDictionary-valued fields can instead be converted into more (potentially nested) param-classes. \n\n## A Note on Parametrization \n\nHdl21 `Generators` have parameters. `Modules` do not. \n\nThis is a deliberate decision, which in this sense makes `hdl21.Module` less feature-rich than the analogous `module` concepts in existing HDLs (Verilog, VHDL, and even SPICE). These languages support what might be called "static parameters" - relatively simple relationships between parent and child-module parameterization. Setting, for example, the width of a signal or number of instances in an array is straightforward. But more elaborate parametrization-cases are either highly cumbersome or altogether impossible to create. (As an example, try using Verilog parametrization to make a programmable-depth binary tree.) Hdl21, in contrast, exposes all parametrization to the full Python-power of its generators. \n\n## Primitives and External Modules\n\nThe leaf-nodes of each hierarchical Hdl21 circuit are generally defined in one of two places: \n\n* `Primitive` elements, typically defined by simulation tools and other EDA software. These include transistors, resistors, capacitors, and other irreducible components. Hdl21 defines a library of such elements in its `hdl21.primitives` module.\n* Outside Hdl21, via `ExternalModules`\n\nHdl21 `Primitives` come in *ideal* and *physical* flavors. The difference is most frequently relevant for passive elements, which can represent either (a) technology-specific passives, e.g. a MIM or MOS capacitor, or (b) an *ideal*  capacitor. Some element-types have solely physical implementations, some are solely ideal, and others include both. A summary of the `hdl21.Primitives` and their flavors:\n\n| Physical             | Ideal            | Alias(es)         | \n| ------------------   | ---------------- | ----------------- | \n| `PhysicalResistor`   | `IdealResistor`  | `R`, `Res`, `Resistor`  | \n| `PhysicalInductor`   | `IdealInductor`  | `L`, `Ind`, `Inductor`  | \n| `PhysicalCapacitor`  | `IdealCapacitor` | `C`, `Cap`, `Capacitor` | \n| `PhysicalShort`      | `IdealShort`     | `Short`                 | \n|                      | `VoltageSource`  | `Vsrc`, `V`             | \n|                      | `CurrentSource`  | `Isrc`, `I`             | \n| `Mos`                |                  |                         | \n| `Diode`              |                  | `D`                     | \n\nAlternately Hdl21 includes an `ExternalModule` type which defines the interface to a module-implementation outside Hdl21. These external definitions are common for instantiating technology-specific modules and libraries. (They might analogously be called *black boxes*.) An example `ExternalModule`: \n\n```python\nimport hdl21 as h \n\n@h.paramclass \nclass BandGapParams:\n    self_destruct = h.Param(\n        dtype=bool, \n        desc="Whether to include the self-destruction feature", \n        default=True\n    )\n\nBandGap = h.ExternalModule(\n    name="BandGap",\n    desc="Example ExternalModule, defined outside Hdl21",\n    port_list=[h.Port("vref"), h.Port("enable")],\n    paramtype=BandGapParams\n)\n```\n\nBoth `Primitives` and `ExternalModules` have names, ordered `Ports`, and a few other pieces of metadata, but no internal implementation: no internal signals, and no instances of other modules. Unlike `Modules`, both *do* have parameters. `Primitives` each have an associated `paramclass`, while `ExternalModules` can optionally declare one via their `paramtype` attribute. Their parameter-types are typically limited to scalar numeric types (`int`, `float`) and `str` (a small subset of those possible for `Generators`) - primarily limited by the need to need to provide them to legacy HDLs. \n\n`Primitives` and `ExternalModules` can be instantiated and connectd in all the same styles as `Modules`: \n\n```python\nimport hdl21 as h\n\nparams = BandGapParams(self_destruct=False)  # Watch out there! \n\n@h.module\nclass BandGapPlus:\n    vref, enable = h.Signals(2)\n    bg = BandGap(params)          # Instantiate the `ExternalModule` defined above\n    bg(vref=vref, enable=enable)  # And call to connect it\n    # ... Everything else ...\n\nparams = h.primitives.DiodeParams(isat=1e-15)\n\n@h.module\nclass DiodePlus:\n    p, n = h.Signals(2)\n    d = h.primitives.D(params)(p=p, n=n)  # Instantiate the Primitive Diode\n    # ... Everything else ...\n```\n\n## Process Technologies\n\nDesigning for a specific implementation technology (or "process development kit", or PDK) with Hdl21 can use either of (or a combination of) two routes: \n\n* Instantiate `ExternalModules` corresponding to the target technology. These would commonly include its tech-specific transistor and passive modules, and potentially larger cells, for example from a cell library. \n* Use `hdl21.Primitives`, each of which is designed to be a technology-independent representation of a primitive component. Moving to a particular technology then generally requires passing the design through an `hdl21pdk` converter. \n\nHdl21 PDKs are Python packages which generally include two primary elements: \n\n* (a) A library `ExternalModules` describing the technology\'s cells, and \n* (b) A `compile` conversion-method which transforms a hierarchical Hdl21 tree, mapping generic `hdl21.Primitives` into the tech-specific `ExternalModules`.\n\n\n---\n## Why Use Python?\n\nCustom IC design is a complicated field. Its practitioners have to know a | lot | of | stuff, independent of any programming background. Many have little or no programming experience at all. Python is reknowned for its accessibility to new programmers, largely attributable to its concise syntax, prototyping-friendly execution model, and thriving community. Moreover, Python has also become a hotbed for many of the tasks hardware designers otherwise learn programming for: numerical analysis, data visualization, machine learning, and the like. \n\nHdl21 exposes the ideas they\'re used to - `Modules`, `Ports`, `Signals` - via as simple of a Python interface as it can. `Generators` are just functions. For many, this fact alone is enough to create powerfully reusable hardware.\n\n## Why *Not* Use {X}?\n\nWe know you have plenty of choice when you fly, and appreciate you choosing Hdl21. A few alternatives and how they compare: \n\n### Schematics\n\nGraphical schematics have been the lingua franca of the custom-circuit field for, well, as long as it\'s been around. Most practitioners are most comfortable in this graphical form. (For plenty of circuits, so are Hdl21\'s authors.) Their most obvious limitation is the lack of capacity for programmable manipulation via something like Hdl21 `Generators`. Some schematic-GUI programs attempt to include "embedded scripting", perhaps even in Hdl21\'s own language (Python). We see those GUIs as entombing your programs in their badness. Hdl21 is instead a library, designed to be used by any Python program you like, sharable and runnable by anyone who has Python. (Which is everyone.) \n\n### Netlists (Spice et al)\n\nTake all of the shortcomings listed for schematics above, and add to them an under-expressive, under-specified, ill-formed, incomplete suite of "programming languages", and you\'ve got netlists. Their primary redeeming quality: existing EDA CAD tools take them as direct input. So Hdl21 Modules export netlists of most popular formats instead. \n\n### (System)Verilog, VHDL, other Existing Dedicated HDLs\n\nThe industry\'s primary, 80s-born digital HDLs Verilog and VHDL have more of the good stuff we want here - notably an open, text-based format, and a more reasonable level of parametrization. And they have the desirable trait of being primary input to the EDA industry\'s core tools. They nonetheless lack the levels of programmability present here. And they generally require one of those EDA tools to execute and do, well, much of anything. Parsing and manipulating them is well-reknowned for requiring a high pain tolerance. Again Hdl21 sees these as export formats. \n\n### Chisel\n\nExplicitly designed for digital-circuit generators at the same home as Hdl21 (UC Berkeley), Chisel encodes RTL-level hardware in Scala-language classes. It\'s the closest of the alternatives in spirit to Hdl21. (And it\'s aways more mature.) If you want big, custom, RTL-level circuits - processors, full SoCs, and the like - you should probably turn to Chisel instead. Chisel makes a number of decisions that make it less desirable for custom circuits, and have accordingly kept their designers\' hands-off. \n\nThe Chisel library\'s primary goal is producing a compiler-style intermediate representation (FIRRTL) to be manipulated by a series of compiler-style passes. We like the compiler-style IR (and may some day output FIRRTL). But custom circuits really don\'t want that compiler. The point of designing custom circuits is dictating exactly what comes out - the compiler *output*. The compiler is, at best, in the way. \n\nNext, Chisel targets *RTL-level* hardware. This includes lots of things that would need something like a logic-synthesis tool to resolve to the structural circuits targeted by Hdl21. For example in Chisel (as well as Verilog and VHDL), it\'s semantically valid to perform an operation like `Signal + Signal`. In custom-circuit-land, it\'s much harder to say what that addition-operator would mean. Should it infer a digital adder? Short two currents together? Stick two capacitors in series? \nMany custom-circuit primitives such as individual transistors actively fight the signal-flow/RTL modeling style assumed by the Chisel semantics and compiler. Again, it\'s in the way. Perhaps more important, many of Chisel\'s abstractions actively hide much of the detail custom circuits are designed to explicitly create. Implicit clock and reset signals serve as prominent examples. \n\nAbove all - Chisel is embedded in Scala. It\'s niche, it\'s complicated, it\'s subtle, it requires dragging around a JVM. It\'s not a language anyone would recommend to expert-designer/ novice-programmers for any reason other than using Chisel. For Hdl21\'s goals, Scala itself is Chisel\'s biggest burden.\n\n### Other Fancy Modern HDLs\n\nThere are lots of other very cool hardware-description projects out there which take Hdl21\'s big-picture approach - embedding hardware idioms as a library in a modern programming languare. All focus on logical and/or RTL-level descriptions, unlike Hdl21\'s structural/ custom/ analog focus. We recommend checking them out: \n\n- [SpinalHDL](https://github.com/SpinalHDL/SpinalHDL)\n- [MyHDL](http://www.myhdl.org/)\n- [Migen](https://github.com/m-labs/migen) / [nMigen](https://github.com/m-labs/nmigen)\n- [Magma](https://github.com/phanrahan/magma)\n- [PyMtl](https://github.com/cornell-brg/pymtl) / [PyMtl3](https://github.com/pymtl/pymtl3)\n- [Clash](https://clash-lang.org/) \n\n\n--- \n\n## Development \n\n* Install [Poetry](https://python-poetry.org/docs/) (the cool-kids\' TOML-configured `pip` replacement). Or drag `pip` along if you know how. \n* Clone this repo & navigate to it \n* `poetry install` will install all dependencies\n* `pytest -s` should yield something like: \n\n\n```\n$ pytest -s\n=== test session starts ===\nplatform darwin -- Python 3.8.5, pytest-5.4.3, py-1.10.0, pluggy-0.13.1\nrootdir: /home/Hdl21\ncollected 21 items                                                                    \n\nhdl21/tests/test_hdl21.py .....................\n\n=== 21 passed in 0.27s ===\n```\n\n',
    'author': 'Dan Fritchman',
    'author_email': 'dan@fritch.mn',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
# hdl21 Elaboration 

Defines the primary `elaborate` method used to flesh out an in-memory `Module` or `Generator`. 
Internally defines and uses a number of hierarchical visitor-classes which traverse the hardware hierarchy, 
performing one or more transformation-passes.  
"""

# Std-Lib Imports
import copy
from enum import Enum
from types import SimpleNamespace
from typing import Union, Any, Dict, List, Optional

# PyPi
from pydantic.dataclasses import dataclass

# Local imports
from .connect import connectable
from .module import Module, ExternalModuleCall
from .instance import InstArray, Instance, PortRef
from .primitives import PrimitiveCall
from .interface import Interface, InterfaceInstance
from .signal import (
    Concat,
    PortDir,
    Signal,
    Visibility,
    Slice,
)
from .generator import Generator, GeneratorCall
from .params import _unique_name
from .instantiable import Instantiable


class Context:
    """Elaboration Context"""

    ...  # To be continued!


# Type short-hand for elaborate-able types
Elabable = Union[Module, GeneratorCall]
# (Plural Version)
Elabables = Union[Elabable, List[Elabable], SimpleNamespace]


def elabable(obj: Any) -> bool:
    # Function to test this, since `isinstance` doesn't work for `Union`.
    return isinstance(obj, (Module, Generator, GeneratorCall))


class _Elaborator:
    """ Base Elaborator Class """

    @classmethod
    def elaborate(cls, top, ctx):
        """ Elaboration entry-point. Elaborate the top-level object. """
        return cls(top, ctx).elaborate_top()

    def __init__(self, top: Elabable, ctx: Context):
        self.top = top
        self.ctx = ctx

    def elaborate_top(self):
        """ Elaborate our top node """
        if not isinstance(self.top, Module):
            raise TypeError
        return self.elaborate_module(self.top)

    def elaborate_generator_call(self, call: GeneratorCall) -> Module:
        """ Elaborate a GeneratorCall """
        # Only the generator-elaborator can handle generator calls; default it to error on others.
        raise RuntimeError(f"Invalid call to elaborate GeneratorCall by {self}")

    def elaborate_module(self, module: Module) -> Module:
        """ Elaborate a Module """
        # Required for all passes. Defaults to `NotImplementedError`.
        raise NotImplementedError

    def elaborate_external_module(self, call: ExternalModuleCall) -> ExternalModuleCall:
        """ Elaborate an ExternalModuleCall """
        # Default: nothing to see here, carry on
        return call

    def elaborate_primitive_call(self, call: PrimitiveCall) -> PrimitiveCall:
        """ Elaborate a PrimitiveCall """
        # Default: nothing to see here, carry on
        return call

    def elaborate_interface_instance(self, inst: InterfaceInstance) -> None:
        """ Elaborate an InterfaceInstance """
        # Annotate each InterfaceInstance so that its pre-elaboration `PortRef` magic is disabled.
        inst._elaborated = True

    def elaborate_interface(self, intf: Interface) -> Interface:
        """ Elaborate an Interface """
        # Default: nothing to see here, carry on
        return intf

    def elaborate_instance_array(self, array: InstArray) -> Instantiable:
        """ Elaborate an InstArray """
        # Turn off `PortRef` magic
        array._elaborated = True
        # And visit the Instance's target
        return self.elaborate_instantiable(array._resolved)

    def elaborate_instance(self, inst: Instance) -> Instantiable:
        """ Elaborate a Module Instance. """
        # This version of `elaborate_instantiable` is the "post-generators" version used by *most* passes.
        # The Generator-elaborator is different, and overrides it.

        # Turn off `PortRef` magic
        inst._elaborated = True
        # And visit the Instance's target
        return self.elaborate_instantiable(inst._resolved)

    def elaborate_instantiable(self, of: Instantiable) -> Instantiable:
        # This version of `elaborate_instantiable` is the "post-generators" version used by *most* passes.
        # The Generator-elaborator is different, and overrides it.
        if not of:
            raise RuntimeError(f"Error elaborating undefined Instance-target {of}")
        if isinstance(of, Module):
            return self.elaborate_module(of)
        if isinstance(of, PrimitiveCall):
            return self.elaborate_primitive_call(of)
        if isinstance(of, ExternalModuleCall):
            return self.elaborate_external_module(of)
        raise TypeError

    @staticmethod
    def flatname(segments: List[str], *, avoid: dict, maxlen: int = 511) -> str:
        """ Create a attribute-name merging string-list `segments`, while avoiding all keys in dictionary `avoid`.
        Commonly re-used while flattening  nested objects and while creating explicit attributes from implicit ones. 
        Raises a `RunTimeError` if no such name can be found of length less than `maxlen`. 
        The default max-length is 511 characters, a value representative of typical limits in target EDA formats. """

        # The default format and result is of the form "_seg0_seg1_".
        # If that is (somehow) in the avoid-keys, append and prepend underscores until it's not.
        name = "_" + "_".join(segments) + "_"
        while True:
            if len(name) > maxlen:
                msg = f"Could not generate a flattened name for {segments}: (trying {name})"
                raise RuntimeError(msg)
            if name not in avoid:  # Done!
                break
            name = "_" + name + "_"  # Collision; append and prepend underscores
        return name


class GeneratorElaborator(_Elaborator):
    """ Hierarchical Generator Elaborator
    Walks a hierarchy from `top` calling Generators. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generator_calls = dict()  # GeneratorCalls to their (Module) results
        self.modules = dict()  # Module ids to references
        self.primitive_calls = dict()  # PrimitiveCall ids to references
        self.ext_module_calls = dict()  # PrimitiveCall ids to references

    def elaborate_top(self):
        """ Elaborate our top node """
        if isinstance(self.top, Module):
            return self.elaborate_module(self.top)
        if isinstance(self.top, GeneratorCall):
            return self.elaborate_generator_call(self.top)
        msg = f"Invalid Elaboration top-level {self.top}, must be a Module or Generator"
        raise TypeError(msg)

    def elaborate_generator_call(self, call: GeneratorCall) -> Module:
        """ Elaborate Generator-function-call `call`. Returns the generated Module. """

        # First check out cache
        if call in self.generator_calls:  # Already done!
            # Give the `call` a reference to its result.
            # Note this *has not* necessarily already happened, as the `self.generator_calls` key may be an equally-valued (but distinct) `GeneratorCall`.
            result = self.generator_calls[call]
            call.result = result
            return result

        # The main event: Run the generator-function
        if call.gen.usecontext:
            m = call.gen.func(call.arg, self.ctx)
        else:
            m = call.gen.func(call.arg)

        # Type-check the result
        # Generators may return other (potentially nested) generator-calls; unwind any of them
        while isinstance(m, GeneratorCall):
            # Note this should hit Python's recursive stack-check if it doesn't terminate
            m = self.elaborate_generator_call(m)
        # Ultimately they've gotta resolve to Modules, or they fail.
        if not isinstance(m, Module):
            raise TypeError(
                f"Generator {call.gen.func.__name__} returned {type(m)}, must return Module."
            )

        # Give the GeneratorCall a reference to its result, and store it in our local dict
        call.result = m
        self.generator_calls[call] = m
        # Create a unique name
        # If the Module that comes back is anonymous, start by giving it a name equal to the Generator's
        if m.name is None:
            m.name = call.gen.func.__name__
        # Then add a unique suffix per its parameter-values
        m.name += "(" + _unique_name(call.arg) + ")"
        # Update the Module's `pymodule`, which generally at this point is `hdl21.generator`
        m._pymodule = call.gen.pymodule
        # And elaborate the module
        return self.elaborate_module(m)

    def elaborate_module(self, module: Module) -> Module:
        """ Elaborate Module `module`. 
        Primarily performs flattening of Instance Arrays, 
        and re-connecting to the resultant flattened instances  """
        if id(module) in self.modules:  # Already done!
            return module

        if not module.name:
            msg = f"Anonymous Module {module} cannot be elaborated (did you forget to name it?)"
            raise RuntimeError(msg)

        # Flatten Instance arrays
        while module.instarrays:
            name, array = module.instarrays.popitem()
            module.namespace.pop(name)
            # Visit the array's target
            target = self.elaborate_instance_array(array)

            # And do the real work: flattening it.
            if array.n < 1:
                raise RuntimeError(f"Invalid InstArray {array} with size {array.n}")
            # Create the new, flat Instances
            new_insts = []
            for k in range(array.n):
                name = self.flatname(
                    segments=[array.name, str(k)], avoid=module.namespace
                )
                inst = module.add(Instance(of=target, name=name))
                new_insts.append(inst)
            # And connect them
            for portname, conn in array.conns.items():
                if isinstance(conn, InterfaceInstance):
                    # All new instances get the same InterfaceInstance
                    for inst in new_insts:
                        inst.connect(portname, conn)
                elif isinstance(conn, (Signal, Slice, Concat)):
                    # Get the target-module port, particularly for its width
                    port = target.get(portname)
                    if not isinstance(port, Signal):
                        msg = f"Invalid port connection of `{portname}` {port} to {conn} in InstArray {array}"
                        raise RuntimeError(msg)

                    if port.width == conn.width:
                        # All new instances get the same signal
                        for inst in new_insts:
                            inst.connect(portname, conn)
                    elif port.width * array.n == conn.width:
                        # Each new instance gets a slice, equal to its own width
                        for k, inst in enumerate(new_insts):
                            # FIXME: will be impacted by slice-indices changes!
                            slize = conn[k * port.width : (k + 1) * port.width - 1]
                            if slize.width != port.width:
                                msg = f"Width mismatch connecting {slize} to {port}"
                                raise RuntimeError(msg)
                            inst.connect(portname, slize)
                    else:  # All other width-values are invalid
                        msg = f"Invalid connection of {conn} of width {conn.width} to port {portname} on Array {array.name} of width {port.width}. "
                        msg += f"Valid widths are either {port.width} (broadcasting across instances) and {port.width * array.n} (individually wiring to each)."
                        raise RuntimeError(msg)
                else:
                    msg = f"Invalid connection to {conn} in InstArray {array}"
                    raise TypeError(msg)

        # Depth-first traverse instances, ensuring their targets are defined
        for inst in module.instances.values():
            self.elaborate_instance(inst)
        # Also visit interface instances, turning off their pre-elab magic
        for intf in module.interfaces.values():
            self.elaborate_interface_instance(intf)

        # Store a reference to the now-expanded Module in our cache, and return it
        self.modules[id(module)] = module
        return module

    def elaborate_instance(self, inst: Instance) -> Instantiable:
        """ Elaborate a Module Instance. """
        # This version differs from `Elaborator` in operating on the *unresolved* attribute `inst.of`,
        # instead of the resolved version `inst._resolved`.

        # Turn off `PortRef` magic
        inst._elaborated = True
        # And visit the Instance's target
        return self.elaborate_instantiable(inst.of)

    def elaborate_instantiable(self, of: Instantiable) -> Instantiable:
        """ Elaborate an Instance target. Adds the capacity to call `GeneratorCall`s to the more-common base-case. """
        if isinstance(of, GeneratorCall):
            return self.elaborate_generator_call(call=of)
        return super().elaborate_instantiable(of)


class ImplicitSignals(_Elaborator):
    """ Explicitly create any implicitly-defined `Signal`s, 
    e.g. those defined by port-to-port connections. """

    def elaborate_module(self, module: Module) -> Module:
        """Elaborate Module `module`. First depth-first elaborates its Instances,
        before creating any implicit Signals and connecting them."""

        # FIXME: much of this can and should be shared with `ImplicitInterfaces`

        # Interfaces must be flattened before this point.
        # Throw an error if not.
        if len(module.interfaces):
            raise RuntimeError(
                f"ImplicitSignals elaborator invalidly invoked on Module {module} with Interfaces {module.interfaces}"
            )

        # Depth-first traverse instances, ensuring their targets are defined
        for inst in module.instances.values():
            self.elaborate_instance(inst)

        # Now work through expanding any implicit-ish connections, such as those from port to port.
        # Start by building an adjacency graph of every `PortRef` that's been instantiated
        portconns = dict()
        for inst in module.instances.values():
            for port, conn in inst.conns.items():
                if isinstance(conn, PortRef):
                    internal_ref = PortRef(inst, port)
                    portconns[internal_ref] = conn

        # Now walk through them, assigning each set to a net
        nets: List[List[PortRef]] = list()
        while portconns:
            # Keep both a list for consistent ordering (we think), and a set for quick membership tests
            this_net = SetList()
            # Grab a random pair
            inner, outer = portconns.popitem()
            this_net.add(inner)
            this_net.add(outer)
            # If it's connected to others, grab those
            while outer in portconns:
                outer = portconns.pop(outer)
                this_net.add(outer)
            # Once we find an object not in `portconns`, we've covered everything connected to the net.
            nets.append(this_net.order)

        # For each net, find and/or create a Signal to replace all the PortRefs with.
        for net in nets:
            # Check whether any of them are connected to declared Signals.
            # And if any are, make sure there's only one
            sig = None
            for portref in net:
                portconn = portref.inst.conns.get(portref.portname, None)
                if isinstance(portconn, Signal):
                    if sig is not None and portconn is not sig:
                        # Ruh roh! shorted between things
                        raise RuntimeError(
                            f"Invalid connection to {portconn}, shorting {[(p.inst.name, p.portname) for p in net]}"
                        )
                    sig = portconn
            # If we didn't find any, go about naming and creating one
            if sig is None:
                # Find a unique name for the new Signal.
                # The default name template is "_inst1_port1_inst2_port2_inst3_port3_"
                signame = self.flatname(
                    segments=[f"{p.inst.name}_{p.portname}" for p in net],
                    avoid=module.namespace,
                )
                # Create the Signal, looking up all its properties from the last Instance's Module
                # (If other instances are inconsistent, later stages will flag them)
                lastmod = portref.inst._resolved
                sig = lastmod.ports.get(portref.portname, None)
                if sig is not None:  # Clone it, and remove any Port-attributes
                    sig = copy.copy(sig)
                    sig.vis = Visibility.INTERNAL
                    sig.direction = PortDir.NONE
                else:
                    raise RuntimeError(
                        f"Invalid port {portref.portname} on Instance {portref.inst.name} in Module {module.name}"
                    )
                # Rename it and add it to the Module namespace
                sig.name = signame
                module.add(sig)
            # And re-connect it to each Instance
            for portref in net:
                portref.inst.conns[portref.portname] = sig

        return module


@dataclass
class FlatInterfaceInst:
    """ Flattened Hierarchical Interface, resolved to constituent Signals """

    inst_name: str  # Interface Instance name
    src: Interface  # Source/ Original Interface
    # Flattened signals-dict, keyed by *original* signal name
    signals: Dict[str, Signal]


class InterfaceFlattener(_Elaborator):
    """ Interface-Flattening Elaborator Pass """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.modules = dict()
        self.replacements = dict()  # Interface replacement-dicts by Module (id)

    def elaborate_module(self, module: Module) -> Module:
        """ Flatten Module `module`s Interfaces, replacing them with newly-created Signals.
        Reconnect the flattened Signals to any Instances connected to said Interfaces. """
        if id(module) in self.modules:
            return module  # Already done!

        # Depth-first traverse our instances first.
        # After this loop, each Instance's Module can have newly-expanded ports -
        # but will not have modified its `conns` dict.
        for inst in module.instances.values():
            self.elaborate_instance(inst)

        # Start a dictionary of interface-replacements
        replacements = dict()

        while module.interfaces:
            # Remove the interface-instance from the module
            name, intf = module.interfaces.popitem()
            module.namespace.pop(name)
            # Flatten it
            flat = self.flatten_interface(intf)
            replacements[name] = flat

            # And add each flattened Signal
            for sig in flat.signals.values():
                signame = self.flatname([intf.name, sig.name], avoid=module.namespace)
                if intf.port:
                    vis_ = Visibility.PORT
                    if intf.role is None:
                        dir_ = PortDir.NONE
                    elif intf.role == sig.src:
                        dir_ = PortDir.OUTPUT
                    elif intf.role == sig.dest:
                        dir_ = PortDir.INPUT
                    else:
                        dir_ = PortDir.NONE
                else:
                    vis_ = Visibility.INTERNAL
                    dir_ = PortDir.NONE
                # Apply all these attributes to our new Signal
                sig.name = signame
                sig.vis = vis_
                sig.direction = dir_
                # And add it to the Module namespace
                module.add(sig)

            # Replace connections to any connected instances
            for inst in module.instances.values():
                for portname in list(inst.conns.keys()):
                    if inst.conns[portname] is intf:
                        inst_replacements = self.replacements[id(inst._resolved)]
                        inst_flat = inst_replacements[portname]
                        for flatname, flatsig in flat.signals.items():
                            inst_signame = inst_flat.signals[flatname].name
                            inst.conns[inst_signame] = flatsig
                        # And remove the connection to the original Interface(s)
                        inst.conns.pop(portname)

            # Re-connect any PortRefs it has given out, as in the form:
            # i = SomeInterface()    # Theoretical Interface with signal-attribute `s`
            # x = SomeModule(s=i.s)  # Connects `PortRef` `i.s`
            # FIXME: this needs to happen for hierarchical interfaces too
            for portref in intf.portrefs.values():
                flatsig = flat.signals.get(portref.portname, None)
                if flatsig is None:
                    raise RuntimeError(
                        f"Port {portref.portname} not found in Interface {intf.of.name}"
                    )
                # Walk through our Instances, replacing any connections to this `PortRef` with `flatsig`
                for inst in module.instances.values():
                    for portname, conn in inst.conns.items():
                        if conn is portref:
                            inst.conns[portname] = flatsig

        self.modules[id(module)] = module
        self.replacements[id(module)] = replacements
        return module

    def flatten_interface(self, intf: InterfaceInstance) -> FlatInterfaceInst:
        """ Convert nested Interface `intf` into a flattened `FlatInterfaceInst` of scalar Signals. 
        Flattening and the underlying Signal-cloning is applied to each Interface *instance*, 
        so each Module-visit need not re-clone the signals of the returned `FlatInterfaceInst`. """

        # Create the flattened version, initializing it with `intf`s scalar Signals
        flat = FlatInterfaceInst(
            inst_name=intf.name, src=intf.of, signals=copy.deepcopy(intf.of.signals)
        )
        # Depth-first walk any interface-instance's definitions, flattening them
        for i in intf.of.interfaces.values():
            iflat = self.flatten_interface(i)
            for sig in iflat.signals.values():
                # Rename the signal, and store it in our flat-intf
                signame = self.flatname([i.name, sig.name], avoid=flat.signals)
                sig.name = signame
                flat.signals[signame] = sig
        return flat


class ImplicitInterfaces(_Elaborator):
    """ Create explicit `InterfaceInstance`s for any implicit ones, 
    i.e. those created through port-to-port connections. """

    def elaborate_module(self, module: Module) -> Module:
        """ Elaborate Module `module`. First depth-first elaborates its Instances,
        before creating any implicit `InterfaceInstance`s and connecting them. """

        # FIXME: much of this can and should be shared with `ImplicitSignals`

        # Depth-first traverse instances, ensuring their targets are defined
        for inst in module.instances.values():
            self.elaborate_instance(inst)

        # Now work through expanding any implicit-ish connections, such as those from port to port.
        # Start by building an adjacency graph of every interface-valued `PortRef` that's been instantiated.
        portconns = dict()
        for inst in module.instances.values():
            for port, conn in inst.conns.items():
                if isinstance(conn, PortRef) and isinstance(
                    conn.inst._resolved, (Module, Interface)
                ):
                    other_port = conn.inst._resolved.get(conn.portname)
                    if other_port is None:
                        raise RuntimeError
                    if isinstance(other_port, InterfaceInstance):
                        internal_ref = PortRef(inst, port)
                        portconns[internal_ref] = conn

        # Now walk through them, assigning each set to a net
        nets: List[List[PortRef]] = list()
        while portconns:
            # Keep both a list for consistent ordering (we think), and a set for quick membership tests
            this_net = SetList()
            # Grab a random pair
            inner, outer = portconns.popitem()
            this_net.add(inner)
            this_net.add(outer)
            # If it's connected to others, grab those
            while outer in portconns:
                outer = portconns.pop(outer)
                this_net.add(outer)
            # Once we find an object not in `portconns`, we've covered everything connected to the net.
            nets.append(this_net.order)

        # For each net, find and/or create a Signal to replace all the PortRefs with.
        for net in nets:
            # Check whether any of them are connected to declared Signals.
            # And if any are, make sure there's only one
            sig = None
            for portref in net:
                portconn = portref.inst.conns.get(portref.portname, None)
                if isinstance(portconn, InterfaceInstance):
                    if sig is not None and portconn is not sig:
                        # Ruh roh! shorted between things
                        raise RuntimeError(
                            f"Invalid connection to {portconn}, shorting {[(p.inst.name, p.portname) for p in net]}"
                        )
                    sig = portconn
            # If we didn't find any, go about naming and creating one
            if sig is None:
                # Find a unique name for the new Signal.
                # The default name template is "_inst1_port1_inst2_port2_inst3_port3_"
                signame = self.flatname(
                    segments=[f"{p.inst.name}_{p.portname}" for p in net],
                    avoid=module.namespace,
                )
                # Create the Signal, looking up all its properties from the last Instance's Module
                # (If other instances are inconsistent, later stages will flag them)
                lastmod = portref.inst._resolved
                sig = lastmod.interface_ports.get(portref.portname, None)
                if sig is not None:
                    sig = copy.copy(sig)
                    sig.port = False
                    sig.role = None
                else:
                    raise RuntimeError(
                        f"Invalid port {portref.portname} on Instance {portref.inst.name} in Module {module.name}"
                    )
                # Rename it and add it to the Module namespace
                sig.name = signame
                module.add(sig)
            # And re-connect it to each Instance
            for portref in net:
                portref.inst.conns[portref.portname] = sig

        return module


class InterfaceConnTypes(_Elaborator):
    """ Check for connection-type-validity on each Interface-valued connection. 
    Note this stage *does not* perform port-direction checking or modification. 
    """

    def elaborate_module(self, module: Module) -> Module:
        """ Check each Instance's connections in `module` """

        # Depth-first traverse instances, checking them first.
        for inst in module.instances.values():
            self.elaborate_instance(inst)

        # And then check each Instance
        for inst in module.instances.values():
            self.check_instance(module, inst)

        # No errors means it checked out, return the Module unchanged
        return module

    def check_instance(self, module: Module, inst: Instance) -> None:
        """ Check the connections of `inst` in parent `module` """

        # Get the Instance's resolve target, and its interface-valued IO
        targ = inst._resolved
        if hasattr(targ, "interface_ports"):
            io = copy.copy(targ.interface_ports)
        else:
            # Still check on no-interface Instances (e.g. Primitives),
            # Mostly to ensure no *connections* are errantly Interface-valued.
            io = []
        # And filter out the interface-valued connections
        conns = {
            k: v for k, v in inst.conns.items() if isinstance(v, InterfaceInstance)
        }

        for portname, conn in conns.items():
            # Pop the port from our list
            port = io.pop(portname, None)
            if port is None:
                msg = f"Connection to invalid interface port {portname} on {inst.name} in {module.name}"
                raise RuntimeError(msg)

            if (  # Check its Interface type
                not isinstance(conn, InterfaceInstance)
                or not isinstance(port, InterfaceInstance)
                or conn.of is not port.of
            ):
                msg = f"Invalid Interface Connection between {conn} and {port} on {inst.name} in {module.name}"
                raise RuntimeError(msg)

        if len(io):  # Check for anything lef over
            msg = f"Unconnected Interface Port(s) {io} on {inst.name} in {module.name}"
            raise RuntimeError(msg)


class SignalConnTypes(_Elaborator):
    """ Check for connection-type-validity between each Instance and its connections. 
    "Connection type validity" includes: 
    * All connections must be Signal-like or Interface-like
    * Signal-valued ports must be connected to Signals of the same width
    * Interface-valued ports must be connected to Interfaces of the same type 
    Note this stage may be run after Interfaces have been flattened, in which case the Interface-checks perform no-ops. 
    They are left in place nonetheless. 
    Note this stage *does not* perform port-direction checking or modification. 
    """

    def elaborate_module(self, module: Module) -> Module:
        """ Check each Instance's connections in `module` """

        # Depth-first traverse instances, checking them first.
        for inst in module.instances.values():
            self.elaborate_instance(inst)

        # And then check each Instance
        for inst in module.instances.values():
            self.check_instance(module, inst)

        # No errors means it checked out, return the Module unchanged
        return module

    def check_instance(self, module: Module, inst: Instance) -> None:
        """ Check the connections of `inst` in parent `module` """

        # Get the Instance's resolve target, and its IO
        targ = inst._resolved
        # FIXME: make this `io` method Module-like-wide
        io = copy.copy(targ.ports)
        if hasattr(targ, "interface_ports"):
            io.update(copy.copy(targ.interface_ports))

        for portname, conn in inst.conns.items():
            # Pop the port from our list
            port = io.pop(portname, None)
            if port is None:
                msg = f"Connection to invalid port {portname} on {inst.name} in {module.name}"
                raise RuntimeError(msg)
            # Checks
            if isinstance(conn, InterfaceInstance):
                # Identity-check the target Interface
                if conn.of is not port.of:
                    msg = f"Invalid Interface Connection between {conn} and {port} on {inst.name} in {module.name}"
                    raise RuntimeError(msg)
            elif connectable(conn):
                # For scalar Signals, check that widths match up
                if conn.width != port.width:
                    msg = f"Invalid Connection between {conn} of width {conn.width} and {port} of width {port.width} on {inst.name} in {module.name}"
                    raise RuntimeError(msg)
            else:
                msg = f"Invalid connection {conn} on {inst.name} in {module.name}"
                raise TypeError(msg)

        if len(io):  # Check for anything lef over
            raise RuntimeError(f"Unconnected IO {io} on {inst.name} in {module.name}")


class SliceResolver(_Elaborator):
    """ Elaboration pass to resolve slices and concatenations to concrete signals. 
    Modifies connections to any nested slices, nested concatenations, or combinations thereof. """

    def elaborate_module(self, module: Module) -> Module:
        # Depth-first traverse instances, covering their targets first.
        for inst in module.instances.values():
            self.elaborate_instance(inst)

        # Then do the real work, updating any necessary connections on each instance
        for inst in module.instances.values():
            self.update_instance(module, inst)

        # No errors means it checked out, return the Module unchanged
        return module

    def update_instance(self, module: Module, inst: Instance) -> None:
        """ Update all `Slice` and `Concat` valued connections to remove nested `Slice`s """
        for portname in inst.conns.keys():
            conn = inst.conns[portname]

            if isinstance(conn, Slice):
                # Resolve the slice-parent(s) to concrete signals
                inst.conns[portname] = _resolve_slice(conn)

            elif isinstance(conn, Concat):
                # Resolve the concatenation to flat parts
                inst.conns[portname] = _resolve_concat(conn)

            # All other connection-types (Signals, Interfaces) are fine


def _list_slice(slize: Slice) -> List[Slice]:
    """ Internal recursive helper for `resolve_slice`. 
    Returns a list of Slices in which each element has a concrete Signal for its parent. """

    # FIXME: some combination of this and elaboration needs to recognize "full slices", at least in naming
    # e.g. `sig[:]`'s name for a width-one signal should resolve to `sig`, not `sig_0`

    if isinstance(slize.signal, Signal):
        # Already all good! Just make a one-element list.
        return [slize]

    if isinstance(slize.signal, Concat):
        # Slice a concatenation.
        # Trick here is we make a list of single-bit slices, then use Python-native slicing-syntax into lists to subset it.
        # Creating those single-bit slices generally requires recursive calls into this method.
        bits = list()
        for part in slize.signal.parts:
            for bitnum in range(0, part.width):
                bits.extend(_list_slice(part[bitnum]))

        # Now python-slice into that list
        # FIXME: #1 Python-style slicing will remove the `+1`. Also, whether abs(step) is the best idea here
        return bits[slize.bot : slize.top + 1 : abs(slize.step)]

    if isinstance(slize.signal, Slice):
        if slize.step < 0:
            raise NotImplementedError("Nested negative slice steps not yet supported")

        # Recursively peel off a bit at a time.
        if slize.width == 1:
            # Base case: slice is one-bit wide. Reach into the parent signal and grab that bit.
            parent = slize.signal  # Note this is also a Slice
            return _list_slice(parent.signal[parent.bot + slize.bot])
        # Otherwise recurse in something like a "cons" pattern, splitting between the first bit and the rest.
        return _list_slice(slize.signal[slize.bot]) + _list_slice(
            slize.signal[slize.bot + slize.step : slize.top : slize.step]
        )

    raise TypeError(f"Invalid attempt to resolve slicing on {slize}")


def _resolve_slice(slize: Slice) -> Union[Slice, Concat]:
    """ Resolve a `Slice` to one or more with "conrete" `Signal`s as parents. 
    
    Slices of other Slices and Slices of Concats are both valid design-time constructions. 
    For example: 
    ```python
    h.Concat(sig1, sig2, sig3)[1] # Slice of a Concat
    sig4[0:2][1] # Slice of a Slice
    ```

    While these may not frequently be created by designers, they are (at least) often created by array broadcasting. 
    As some point their parents must be resolved to their original Signals, at minimum before export-level name resolution. 

    Resolving Concatenations can generally resolve to more than one Slice, as in: 
    ```python
    h.Concat(sig1[0], sig2[0], sig3[0])[0:1] # Requires slices of `sig1` and `sig2`
    ``` 
    Such cases create and return a Concatenation. """

    ls = _list_slice(slize)
    if len(ls) == 1:  # Resolved to single parent-Signal
        return ls[0]
    elif len(ls) > 1:  # Multiple parts required - concatenate them
        return Concat(*ls)

    raise RuntimeError(f"Error resolving Slice {slize}")


def _resolve_concat(conc: Concat) -> Concat:
    """ Resolve a Concatenation into (a) concrete Signals and (b) Slices of concrete Signals. 
    Removes nested concatenations and resolves slices along the way. """

    if not len(conc.parts):
        raise RuntimeError("Concatenation with no parts")

    if all(_flat_concatable(p) for p in conc.parts):
        return conc  # Already all good!

    if isinstance(conc.parts[0], Concat):
        # Recursively cover the first element, and all others
        first = _resolve_concat(conc.parts[0])
        rest = _resolve_concat(Concat(*conc.parts[1:]))
        return Concat(*(first.parts + rest.parts))

    if isinstance(conc.parts[0], Slice):
        # Resolve everything within the Slice to a list of concrete-Signal slices
        first = _resolve_slice(conc.parts[0])
        # Pass everything else recursively back to this method
        rest = _resolve_concat(Concat(*conc.parts[1:]))
        # And concatenate the two
        return Concat(*(first + rest.parts))

    # Otherwise peel off as many Signals and concrete-Signal Slices as we can
    for idx in range(len(conc.parts)):
        if _flat_concatable(conc.parts[idx]):
            continue
        # Hit our first "compound" entry. Split the list here.
        first = conc.parts[:idx]
        rest = _resolve_concat(Concat(*conc.parts[idx:]))
        return Concat(*(first + rest.parts))

    raise RuntimeError("Unable to resolve concatenation")


def _flat_concatable(s: Union[Signal, Slice, Concat]) -> bool:
    """ Boolean indication of whether `s` is suitable for flattened Concatenations. 
    Such objects must be either: 
    * (a) A Signal, or
    * (b) A Slice into a Signal
    Notable exceptions include Concats and nested Slices of Concats and other Slices. """
    return isinstance(s, Signal) or (
        isinstance(s, Slice) and isinstance(s.signal, Signal)
    )


class SetList:
    """ A common combination of a hash-set and ordered list of the same items. 
    Used for keeping ordered items while maintaining quick membership testing. """

    def __init__(self):
        self.set = set()
        self.list = list()

    def __contains__(self, item):
        return item in self.set

    def add(self, item):
        if item not in self.set:
            self.set.add(item)
            self.list.append(item)

    @property
    def order(self):
        return self.list


class ElabPass(Enum):
    """ Enumerated Elaborator Passes
    Each has a `value` attribute which is an Elaborator-pass, 
    and a `name` attribute which is a (Python-enum-style) capitalized name. 
    Specifying  """

    RUN_GENERATORS = GeneratorElaborator
    IMPLICIT_INTERFACES = ImplicitInterfaces
    INTERFACE_CONN_TYPES = InterfaceConnTypes
    FLATTEN_INTERFACES = InterfaceFlattener
    IMPLICIT_SIGNALS = ImplicitSignals
    SIGNAL_CONN_TYPES = SignalConnTypes
    RESOLVE_SLICES = SliceResolver

    @classmethod
    def default(cls) -> List["ElabPass"]:
        """ Return the default ordered Elaborator Passes. """
        # Returns each in definition order
        return list(ElabPass)


def elab_all(top: Elabables, **kwargs) -> List[Elabable]:
    """ Elaborate everything we can find - potentially recursively - in `Elabables` `top`. 

    Results are returned in a list, not necessarily reproducing the structure of `top`. 
    Note the *attributes* of `top` are also generally modified in-place, allowing access to their elaboration results. """
    # Recursively create a list of all elab-able types in `obj`
    ls = []
    _list_elabables_helper(top, ls)
    # Elaborate each, and return them as a list
    return [elaborate(top=t, **kwargs) for t in ls]


def _list_elabables_helper(obj: Any, accum: list) -> None:
    """ Recursive helper for hierarchically finding elaborate-able things in `obj`. 
    Newly-found items are appended to accumulation-list `accum`. """
    if elabable(obj):
        accum.append(obj)
    elif isinstance(obj, list):
        [_list_elabables_helper(i, accum) for i in obj]
    elif isinstance(obj, SimpleNamespace):
        # Note this skips over non-elaboratable items (e.g. names), where the list demands all be suitable.
        for i in obj.__dict__.values():
            if isinstance(i, (SimpleNamespace, list)) or elabable(i):
                _list_elabables_helper(i, accum)
    else:
        raise TypeError(f"Attempting Invalid Elaboration of {obj}")


def elaborate(
    top: Elabable,
    *,
    ctx: Optional[Context] = None,
    passes: Optional[List[ElabPass]] = None,
) -> Module:
    """ In-Memory Elaboration of Generator or Module `top`. 
    
    Optional `passes` lists the ordered `ElabPass`es to run. By default it runs the order specified by `ElabPass.default`. 
    Note the order of passes is important; many depend upon others to have completed before they can successfully run. 

    Optional `Context` field `ctx` is not yet supported. 

    `elaborate` executes elaboration of a *single* `top` object. 
    For (plural) combinations of `Elabable` objects, use `elab_all`. 
    """
    # Expand default values
    ctx = ctx or Context()
    passes = passes or ElabPass.default()

    # Pass `top` through each of our passes, in order
    res = top
    for pass_ in passes:
        res = pass_.value.elaborate(top=res, ctx=ctx)
    return res



""" 
# Hdl21 Hierarchical Instances 

Create instances of Modules, Generators, and Primitives in a hierarchy
"""

from pydantic.dataclasses import dataclass
from typing import Optional, Union

# Local imports
from .connect import connectable, connects


@connectable
@dataclass
class PortRef:
    """ Port Reference to an Instance """

    inst: Union["Instance", "InstArray", "InterfaceInstance"]
    portname: str

    def __eq__(self, other) -> bool:
        """ Port-reference equality requires *identity* between instances 
        (and of course equality of port-name). """
        return self.inst is other.inst and self.portname == other.portname

    def __hash__(self):
        """ Hash references as the tuple of their instance-address and name """
        return hash((id(self.inst), self.portname))


@connects
@dataclass
class InstanceRef:
    """ Instance Reference from an Array or Group """

    parent: "InstArray"
    name: Union[int, str]

    def __eq__(self, other) -> bool:
        """ Equality requires *identity* between parents 
        (and of course equality of name). """
        return self.parent is other.parent and self.name == other.name

    def __hash__(self):
        """ Hash references as the tuple of their instance-address and name """
        return hash((id(self.parent), self.name))


class _Instance:
    """ Shared base class for Instance-like types (Instance, InstArray) """

    def __init__(
        self, of: "Instantiable", *, name: Optional[str] = None,
    ):
        from .instantiable import _is_instantiable

        if not _is_instantiable(of):
            raise RuntimeError(f"Invalid instance of {of}")

        self.name = name
        self.of = of
        self.conns = dict()
        self.portrefs = dict()
        self._parent_module = None  # Instantiating module
        self._elaborated = False
        self._initialized = True

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name} of={self.of})"

    @property
    def _resolved(self,) -> Optional["Instantiable"]:
        """ Property to retrieve the Instance's resolved Module, if complete. 
        Returns `None` if unresolved. """
        from .generator import GeneratorCall

        if isinstance(self.of, GeneratorCall):
            return self.of.result
        return self.of


@connects
class Instance(_Instance):
    """ Hierarchical Instance of another Module or Generator """

    _specialcases = [
        "name",
        "of",
        "conns",
        "portref",
        "portrefs",
        "connect",
        "_module",
        "_resolved",
        "_elaborated",
        "_initialized",
    ]

    def _mult(self, other: int) -> "InstArray":
        """ Instance by integer multiplication. 
        Creates an Instance Array of size `other`. """
        if not isinstance(other, int):
            return NotImplemented
        return self._to_array(num=other)

    __mul__ = __rmul__ = _mult  # Apply `_mult` on both left and right

    def _to_array(self, num: int) -> "InstArray":
        """ Create an Instance Array from an Instance """
        # Several contraints asserted here which may eventually be relaxed.
        # * No port-references (yet)
        # * Not a member of a module (yet)
        if len(self.portrefs) > 0:
            msg = f"Cannot convert Instance {self} with outstanding port-references {self.portrefs} to Array"
            raise RuntimeError(msg)
        if self._parent_module is not None:
            msg = f"Cannot convert Instance {self} already inserted in Module {self._parent_module} to Array"
            raise RuntimeError(msg)

        # Checks out. Create the array.
        return InstArray(of=self.of, n=num, name=self.name)(**self.conns)


@connects
class InstArray(_Instance):
    """ Array of `n` Instances """

    _specialcases = [
        "name",
        "of",
        "conns",
        "portref",
        "portrefs",
        "instrefs",
        "connect",
        "_elaborated",
        "_initialized",
        "_module",
    ]

    def __init__(
        self, of: "Instantiable", n: int, name: Optional[str] = None,
    ):
        self.instrefs = dict()
        self.n = n
        super().__init__(of=of, name=name)

    def __getitem__(self, idx: int) -> InstanceRef:
        if not isinstance(idx, int):
            return RuntimeError(f"Illegal indexing into {self} with non-integer {idx}")
        raise NotImplementedError  # FIXME: coming soon!


# Get the runtime type-checking to understand the types forward-referenced and then defined here
from .interface import InterfaceInstance

PortRef.__pydantic_model__.Config.arbitrary_types_allowed = True
PortRef.__pydantic_model__.update_forward_refs()


def calls_instantiate(cls: type) -> type:
    """ Decorator which adds 'calls produce `hdl21.Instance`s' functionality. """

    def __call__(self, **kwargs) -> Instance:
        """ Calls Create `hdl21.Instances`, 
        and pass any (keyword-only) arguments to said `Instances`, 
        generally to connect-by-call. """
        return Instance(of=self)(**kwargs)

    # Check for an existing __call__ method, and if there is one, bail
    if "__call__" in cls.__dict__:
        msg = f"Hdl21 Internal Error: Invalid conflict between `calls_instantiate` decorator and explicit `__call__` method on {cls}"
        raise RuntimeError(msg)
    cls.__call__ = __call__
    return cls



"""
# Hierachy Walker Module 
"""

# Std-Lib
from types import SimpleNamespace

# Local imports
from .module import Module, ExternalModuleCall
from .instance import Instance
from .primitives import PrimitiveCall
from .interface import Interface, InterfaceInstance
from .signal import Port, PortDir, Signal, Visibility
from .generator import Generator, GeneratorCall


class HierarchyWalker:
    """Hierarchical Traversal Helper
    Walks a hierarchical `Module` or `SimpleNamespace` thereof.
    Designed to be used as a base-class for extensions such as
    process-specific instance-replacements."""

    def visit_namespace(self, ns: SimpleNamespace):
        for k, v in ns.__dict__.items():
            if isinstance(v, Module):
                self.visit_module(v)
            if isinstance(v, SimpleNamespace):
                self.visit_namespace(v)

    def visit_generator_call(self, call: GeneratorCall):
        raise NotImplementedError

    def visit_module(self, module: Module):
        for inst in module.instances.values():
            self.visit_instance(inst)
        for inst in module.interfaces.values():
            self.visit_interface_instance(inst)

    def visit_external_module(self, call: ExternalModuleCall):
        return

    def visit_primitive_call(self, call: PrimitiveCall):
        return

    def visit_interface_instance(self, inst: InterfaceInstance):
        return

    def visit_instance(self, inst: Instance):
        if isinstance(inst.of, GeneratorCall):
            return self.visit_generator_call(call=inst.of)
        if isinstance(inst.of, Module):
            return self.visit_module(inst.of)
        if isinstance(inst.of, PrimitiveCall):
            return self.visit_primitive_call(inst.of)
        if isinstance(inst.of, ExternalModuleCall):
            return self.visit_external_module(inst.of)
        raise TypeError(f"Invalid Instance of {inst.of}")



"""
Type-alias for `Instantiable` hdl21 types. Each is valid as the `of` field of `hdl21.Instance`s, 
and thus supports its "connect by call" and "connect by assigment" semantics. 
"""

from typing import Any, Union, get_args

from .module import Module, ExternalModuleCall
from .generator import GeneratorCall
from .primitives import PrimitiveCall


# Instantiable types-union
Instantiable = Union[Module, ExternalModuleCall, GeneratorCall, PrimitiveCall]


def _is_instantiable(val: Any) -> bool:
    """ Boolean indication of whether `val` is an `Instantiable` type. """
    return isinstance(val, get_args(Instantiable))
